package twostackqueue;

import java.util.Random;

/**
 * This program tests two queue implementations (TwoStackQueue and
 * StandardLibraryQueue) for correctness and then does run time testing on both
 * implementations.
 * All results are printed to standard output.
 * This program ignores commmand-line arguments.
 */
public class Main {

  /**
   * isImplementationCorrect(queue) determines if queue correctly implements the
   * Dispenser interface with a first-in-first-out policy.
   * @param dispenser the queue to be tested
   * @return true if it has a valid first-in-first-out policy
   */
  private static boolean isImplementationCorrect(Dispenser<Integer> queue) {
    // Add 100 items.
    for (int i = 0; i < 100; i++) {
      queue.add(i);
    }
    // Do 100 pairs of removes and adds, checking correctness of removes.
    int expected = 0;
    for (int i = 100; i < 200; i++) {
      if (queue.remove() != expected) {
        return false;
      }
      expected++;
      queue.add(i);
    }
    // Do 100 more removes, checking their correctness.
    while (expected < 200) {
      if (queue.remove() != expected) {
        return false;
      }
      expected++;
    }
    // Check if the queue is empty.
    return queue.isEmpty();
  }

  /**
   * printCorrectness(name, queue) prints correctness check results for queue to
   * standard out.
   * The name parameter identifies the queue that is being tested.
   * @param name the name for the queue to be tested
   * @param dispenser the queue to be tested
   */
  private static void printCorrectness(String name, Dispenser<Integer> queue) {
    System.out.print(name + " implementation ");
    if (isImplementationCorrect(queue)) {
      System.out.println("passed!");
    } else {
      System.out.println("failed!");
    }
  }

  /**
   * getRuntime(dispenser, seed) returns the run time in a test on dispenser
   * using seed for random number generation.
   * @param dispenser the dispenser to be tested
   * @param seed the seed for random number generation
   * @return the run time
   */
  private static double getRunTime(Dispenser<Integer> dispenser, int seed) {
    Random rng = new Random(seed);
    // First add a bunch of entries so that the dispenser is extremely unlikely
    // to become empty during the run time test.
    for (int i = 0; i < 100000; i++) {
      dispenser.add(rng.nextInt());
    }
    // Now start the run time test.
    long startTime = System.currentTimeMillis();
    for (int i = 0; i < 10000000; i++) {
      // The random number generator determines whether an item will be added or
      // removed.
      int number = rng.nextInt();
      if (number % 2 == 0) {
        // For an add the random number generator determines the number to be
        // added.
        dispenser.add(rng.nextInt());
      } else {
        dispenser.remove();
      }
    }
    return (System.currentTimeMillis() - startTime) / 1000.0;
  }

  /**
   * printRunTime(name, dispenser, seed) prints run time results for dispenser
   * to standard out.
   * The name parameter identifies the dispenser that is being tested.
   * The seed parameter is used for random number generation.
   * @param name the name for the queue to be tested
   * @param dispenser the dispenser to be tested
   * @param seed the seed for random number generation
   */
  private static void printRunTime(String name, Dispenser dispenser, int seed) {
    dispenser = new TwoStackQueue<Integer>();
    double runTime = getRunTime(dispenser, seed);
    System.out.println(name + " run time is " + runTime);
  }

  /**
   * Main.main(args) executes correctness tests and run time tests on two queue
   * implementations: TwoStackQueue and StandardLibraryQueue.
   * @param args not used
   */
  public static void main(String[] args) {
    // Print correctness results for both implementations.
    printCorrectness("StandardLibraryQueue", new StandardLibraryQueue<Integer>());
    printCorrectness("TwoStackQueue", new TwoStackQueue<Integer>());
    // Do 10 run time tests for each implementation using the same random number
    // generator seed in each pair of tests.
    Random seedGenerator = new Random();
    for (int i = 0; i < 10; i++) {
      int seed = seedGenerator.nextInt();
      printRunTime("StandardLibraryQueue", new StandardLibraryQueue<Integer>(), seed);
      printRunTime("TwoStackQueue", new TwoStackQueue<Integer>(), seed);
    }
  }

}
